import abc

import quizgen.variant

class Converter(abc.ABC):
    def __init__(self, answer_key = False, **kwargs):
        super().__init__()

        self.answer_key = answer_key

    @abc.abstractmethod
    def convert_variant(self, variant, **kwargs):
        pass

    def convert_question(self, question, variant = None, **kwargs):
        """
        Convert a single question.
        If a variant is passed in, the converter will change it.
        """

        if (variant is None):
            variant = quizgen.variant.Variant.get_dummy()

        variant.questions = [question]

        return self.convert_variant(variant, **kwargs)



"""
Convert a quiz into TeX using templates.
"""

import os

import quizgen.constants
import quizgen.converter.template

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-tex')

JINJA_OPTIONS = {
    'block_start_string': '<%',
    'block_end_string': '%>',
    'variable_start_string': '<<',
    'variable_end_string': '>>',
    'comment_start_string': '<#',
    'comment_end_string': '#>',
}

class TexTemplateConverter(quizgen.converter.template.TemplateConverter):
    def __init__(self, template_dir = DEFAULT_TEMPLATE_DIR, **kwargs):
        super().__init__(quizgen.constants.FORMAT_TEX, template_dir, jinja_options = JINJA_OPTIONS, **kwargs)

    def clean_solution_content(self, document):
        tex = document.to_format(quizgen.constants.FORMAT_TEX)
        if ('\\' not in tex):
            return tex

        content = document.to_format(quizgen.constants.FORMAT_TEXT)
        content = content.replace('\\', '\\textbackslash{}')

        return content



import argparse
import sys

import quizgen.constants
import quizgen.log
import quizgen.parser

def run(args):
    document = quizgen.parser.parse_file(args.path)

    content = document.to_format(args.format)
    print(content)

    return 0

def _get_parser():
    parser = argparse.ArgumentParser(description =
        "Parse a single file and output the results of the parse.")

    parser.add_argument('path',
        type = str,
        help = 'The path to parse.')

    parser.add_argument('--format',
        action = 'store', type = str, default = quizgen.constants.FORMAT_JSON,
        choices = quizgen.constants.DOC_FORMATS,
        help = 'Output the parsed document in this format (default: %(default)s).')

    quizgen.log.set_cli_args(parser)

    return parser

def main():
    args = _get_parser().parse_args()
    quizgen.log.init_from_args(args)
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import datetime
import json
import os

import json5

import quizgen.question
import quizgen.common
import quizgen.parser
import quizgen.quiz

DUMMY_DATA = {
    'title': 'Dummy Title',
    'description': 'Dummy description.',
    'course_title': 'Dummy Course',
    'term_title': 'Dummy Term',
    'version': '0.0.0',
}

class Variant(object):
    """
    A quiz varint is an instantiation of a quiz with specific questions chosen for each group.
    Variants no longer have groups, just questions

    Variants should be created from quizzes, and will therefore not be deeply validated,
    just checked for null values.
    """

    def __init__(self,
            title = None,
            course_title = None, term_title = None,
            description = None, description_document = None,
            date = None,
            questions = None,
            version = None, seed = None,
            **kwargs):
        self.title = title
        self.course_title = course_title
        self.term_title = term_title
        self.date = date

        self.description = description
        self.description_document = description_document

        self.questions = questions

        self.version = version
        self.seed = seed

        self.validate()

    def validate(self):
        values = self.__dict__.copy()
        for (key, value) in values.items():
            if (value is None):
                raise quizgen.common.QuizValidationError("Empty variant value: '%s'." % (key))

    def to_dict(self, include_docs = True):
        value = self.__dict__.copy()

        if ('date' in value):
            value['date'] = value['date'].isoformat()

        value['questions'] = [question.to_dict(include_docs = include_docs) for question in self.questions]

        if (include_docs):
            value['description_document'] = self.description_document.to_pod()
        else:
            del value['description_document']

        return value

    @staticmethod
    def from_path(path, override_base_dir = False):
        path = os.path.abspath(path)

        with open(path, 'r') as file:
            data = json5.load(file)

        base_dir = None
        if (override_base_dir):
            base_dir = os.path.dirname(path)

        return Variant.from_dict(data, base_dir = base_dir)

    @staticmethod
    def from_dict(data, base_dir = None):
        data = data.copy()

        if (base_dir is not None):
            data['base_dir'] = base_dir
        elif ('base_dir' not in data):
            data['base_dir'] = '.'

        if ('date' in data):
            data['date'] = datetime.datetime.fromisoformat(data['date'])

        data['description_document'] = quizgen.parser.parse_text(data['description'], base_dir = base_dir)
        data['questions'] = [quizgen.question.Question.from_dict(question, base_dir = base_dir) for question in data['questions']]

        return Variant(**data)

    def to_json(self, indent = 4, include_docs = True):
        return json.dumps(self.to_dict(include_docs = include_docs), indent = indent)

    def num_questions(self):
        return len(self.questions)

    @staticmethod
    def get_dummy():
        """
        Get a "dummy" variant that has no real information.
        """

        return quizgen.quiz.Quiz(**DUMMY_DATA).create_variant()



import glob
import os

import quizgen.converter.htmltemplate
import quizgen.converter.json
import quizgen.converter.textemplate
import quizgen.quiz
import tests.base

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

CONVERTERS = [
    quizgen.converter.json.JSONConverter,
    quizgen.converter.textemplate.TexTemplateConverter,
    quizgen.converter.htmltemplate.HTMLTemplateConverter,
]

class QuizQuestionsTest(tests.base.BaseTest):
    """
    Compile all the good test quiz question into a quiz and test that quiz.
    """

    _quiz = None
    _num_paths = -1

    @classmethod
    def setUpClass(cls):
        good_paths, _ = tests.base.discover_question_tests()

        quiz_info = {
            'title': 'Test Quiz',
            'course_title': "Test Course",
            'term_title': "Test Term",
            'description': 'testing',
            'version': '0.0.0',
            'groups': [],
        }

        for path in good_paths:
            name = os.path.basename(os.path.dirname(path))

            quiz_info['groups'].append({
                'name': name,
                'questions': [path],
            })

        QuizQuestionsTest._quiz = quizgen.quiz.Quiz.from_dict(quiz_info, THIS_DIR)
        QuizQuestionsTest._num_paths = len(good_paths)

    def testNumQuestions(self):
        self.assertEqual(QuizQuestionsTest._num_paths, QuizQuestionsTest._quiz.num_questions())

def _add_converter_tests():
    for converter in CONVERTERS:
        for key in [True, False]:
            for shuffle in [True, False]:
                test_name = 'test_converter__%s__key_%s__shuffle_%s' % (converter.__name__, str(key), str(shuffle))
                setattr(QuizQuestionsTest, test_name, _get_template_test(converter, key, shuffle))

def _get_template_test(converter_class, key, shuffle):
    def __method(self):
        converter = converter_class(answer_key = key)

        variant = QuizQuestionsTest._quiz.create_variant(all_questions = True)
        variant.shuffle_answers = shuffle

        content = converter.convert_variant(variant)
        self.assertTrue(len(content) > 10)

    return __method

_add_converter_tests()



import math
import os
import random
import string

import quizgen.constants
import quizgen.converter.converter
import quizgen.parser
import quizgen.variant

import jinja2

TEMPLATE_FILENAME_QUIZ = 'quiz.template'
TEMPLATE_FILENAME_QUESTION_SEP = 'question-separator.template'

RIGHT_IDS = string.ascii_uppercase
LEFT_IDS = [str(i + 1) for i in range(len(RIGHT_IDS))]

DEFAULT_JINJA_OPTIONS = {
    'trim_blocks': True,
    'lstrip_blocks': True,
    'autoescape': jinja2.select_autoescape(),
}

class TemplateConverter(quizgen.converter.converter.Converter):
    def __init__(self, format, template_dir,
            jinja_options = {}, jinja_filters = {},
            **kwargs):
        super().__init__(**kwargs)

        if (not os.path.isdir(template_dir)):
            raise ValueError("Provided template dir ('%s') does not exist or is not a dir." % (
                template_dir))

        self.format = format
        self.template_dir = template_dir

        self.jinja_options = DEFAULT_JINJA_OPTIONS.copy()
        self.jinja_options.update(jinja_options)

        self.env = jinja2.Environment(
            loader = jinja2.FileSystemLoader(self.template_dir, followlinks = True),
            **self.jinja_options,
        )

        for (name, function) in jinja_filters.items():
            self.env.filters[name] = function

        # Methods to generate answers.
        # Signatuire: func(self, question_index, question_number, question, variant)
        self.answer_functions = {
            quizgen.constants.QUESTION_TYPE_ESSAY: 'create_answers_essay',
            quizgen.constants.QUESTION_TYPE_FIMB: 'create_answers_fimb',
            quizgen.constants.QUESTION_TYPE_FITB: 'create_answers_fitb',
            quizgen.constants.QUESTION_TYPE_MA: 'create_answers_ma',
            quizgen.constants.QUESTION_TYPE_MATCHING: 'create_answers_matching',
            quizgen.constants.QUESTION_TYPE_MCQ: 'create_answers_mcq',
            quizgen.constants.QUESTION_TYPE_MDD: 'create_answers_mdd',
            quizgen.constants.QUESTION_TYPE_NUMERICAL: 'create_answers_numerical',
            quizgen.constants.QUESTION_TYPE_SA: 'create_answers_sa',
            quizgen.constants.QUESTION_TYPE_TEXT_ONLY: 'create_answers_text_only',
            quizgen.constants.QUESTION_TYPE_TF: 'create_answers_tf',
        }

    def convert_variant(self, variant, **kwargs):
        if (not isinstance(variant, quizgen.variant.Variant)):
            raise ValueError("Template quiz converter requires a quizgen.variant.Variant type, found %s." % (type(variant)))

        questions_text = self.create_questions(variant)

        variant_context = variant.to_dict(include_docs = False)
        variant_context['description_text'] = variant.description_document.to_format(self.format)

        context = {
            'quiz': variant_context,
            'answer_key': self.answer_key,
            'questions_text': questions_text,
        }

        variant_template = self.env.get_template(TEMPLATE_FILENAME_QUIZ)
        text = variant_template.render(**context)

        return text

    def create_questions(self, variant):
        questions = []

        question_number = 1
        for question_index in range(len(variant.questions)):
            question = variant.questions[question_index]

            if (question_index != 0):
                questions.append(self.create_question_separator(variant))

            try:
                questions.append(self.create_question(question_index, question_number, question, variant))
            except Exception as ex:
                raise ValueError("Failed to convert question %d (%s)." % (question_index, question.base_name)) from ex

            if (not question.should_skip_numbering()):
                question_number += 1

        return "\n\n".join(questions)

    def create_question(self, question_index, question_number, question, variant):
        question_type = question.question_type
        if (question_type not in self.answer_functions):
            raise ValueError("Unsupported question type: '%s'." % (question_type))

        data = question.to_dict(include_docs = False)
        data['prompt_text'] = question.prompt_document.to_format(self.format)
        data['id'] = question_index
        data['number'] = question_number

        # Stash the old answers and add in new ones.
        data['answers_raw'] = data['answers']
        answers_method = getattr(self, self.answer_functions[question_type])
        data['answers'] = answers_method(question_index, question_number, question, variant)

        context = {
            'quiz': variant,
            'answer_key': self.answer_key,
            'question': data,
        }

        template_name = "%s.template" % (question_type)
        template = self.env.get_template(template_name)
        text = template.render(**context)

        return text

    def clean_solution_content(self, document):
        """
        An opportunity for children to clean the text of a solution before it is entered into a key.
        For example, tex solutions are hacky and cannot use certain functions.
        """

        return document.to_format(self.format)

    def create_question_separator(self, variant):
        context = {
            'quiz': variant,
            'answer_key': self.answer_key,
        }

        template = self.env.get_template(TEMPLATE_FILENAME_QUESTION_SEP)
        text = template.render(**context)

        return text

    def create_answers_tf(self, question_index, question_number, question, variant):
        return question.answers

    def create_answers_matching(self, question_index, question_number, question, variant):
        lefts = []
        rights = []

        # {left_index: right_index, ...}
        matches = {}

        for (left, right) in question.answers_documents['matches']:
            matches[len(lefts)] = len(rights)

            lefts.append(left.to_format(self.format))
            rights.append(right.to_format(self.format))

        for right in question.answers_documents['distractors']:
            rights.append(right.to_format(self.format))

        left_ids = self.get_matching_left_ids()
        right_ids = self.get_matching_right_ids()

        if (len(lefts) > len(left_ids)):
            raise ValueError("Too many left-hand values for a matching question. Found: %d, Max %d." % (len(lefts) > len(left_ids)))

        if (len(rights) > len(right_ids)):
            raise ValueError("Too many right-hand values for a matching question. Found: %d, Max %d." % (len(rights) > len(right_ids)))

        if (question.answers.get('shuffle', False)):
            seed = question.answers.get('shuffle_seed', None)
            rng = random.Random(seed)

            # Shuffle the left and right options while maintining the match mapping.
            left_indexes = list(range(len(lefts)))
            right_indexes = list(range(len(rights)))

            rng.shuffle(left_indexes)
            rng.shuffle(right_indexes)

            new_lefts = [lefts[index] for index in left_indexes]
            lefts = new_lefts

            new_rights = [rights[index] for index in right_indexes]
            rights = new_rights

            new_matches = {left_indexes.index(old_left_index): right_indexes.index(old_right_index) for (old_left_index, old_right_index) in matches.items()}
            matches = new_matches

        # Augment the left and rights with more information for the template.
        for left_index in range(len(lefts)):
            right_index = matches[left_index]

            lefts[left_index] = {
                'id': "%d.%s" % (question_index, left_ids[left_index]),
                'text': lefts[left_index],
                'solution': right_ids[right_index],
                'solution_id': "%d.%s" % (question_index, right_ids[right_index]),
            }

        for right_index in range(len(rights)):
            rights[right_index] = {
                'id': "%d.%s" % (question_index, right_ids[right_index]),
                'text': rights[right_index],
                'label': right_ids[right_index],
            }

        return {
            'lefts': lefts,
            'rights': rights,
            'matches': matches,
        }

    def get_matching_left_ids(self):
        return LEFT_IDS

    def get_matching_right_ids(self):
        return RIGHT_IDS

    def create_answers_mcq(self, question_index, question_number, question, variant):
        return self._create_answers_mcq_list(question.answers, question.answers_documents)

    def _create_answers_mcq_list(self, answers, documents):
        choices = []

        for i in range(len(answers)):
            choices.append({
                'correct': answers[i]['correct'],
                'text': documents[i].to_format(self.format),
            })

        return choices

    def create_answers_text_only(self, question_index, question_number, question, variant):
        return None

    def create_answers_numerical(self, question_index, question_number, question, variant):
        answer = question.answers[0]

        if (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_EXACT):
            if (math.isclose(answer['margin'], 0.0)):
                content = "%s" % (str(answer['value']))
            else:
                content = "%s ± %f" % (str(answer['value']), answer['margin'])
        elif (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_PRECISION):
            content = "[%s, %s]" % (str(answer['min']), str(answer['max']))
        elif (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_RANGE):
            content = "%s (precision: %s)" % (str(answer['value']), str(answer['precision']))
            raise ValueError(f"Unknown numerical answer type: '{answer['type']}'.")

        document = quizgen.parser.parse_text(content)

        return {
            'solution': self.clean_solution_content(document),
            'dirty_solution': document.to_format(self.format),
        }

    def create_answers_mdd(self, question_index, question_number, question, variant):
        answers = []

        for key, values in question.answers.items():
            answers.append({
                'label': question.answers_documents[key]['key'].to_format(self.format),
                'choices': self._create_answers_mcq_list(values, question.answers_documents[key]['values']),
            })

        return answers

    def create_answers_ma(self, question_index, question_number, question, variant):
        return self._create_answers_mcq_list(question.answers, question.answers_documents)

    def create_answers_fimb(self, question_index, question_number, question, variant):
        answers = []

        for key, values in question.answers.items():
            document = question.answers_documents[key]['values'][0]

            answers.append({
                'label': question.answers_documents[key]['key'].to_format(self.format),
                'solution': self.clean_solution_content(document),
                'dirty_solution': document.to_format(self.format),
            })

        return answers

    def create_answers_fitb(self, question_index, question_number, question, variant):
        document = question.answers_documents[""]["values"][0]

        return {
            'solution': self.clean_solution_content(document),
            'dirty_solution': document.to_format(self.format),
        }

    def create_answers_sa(self, question_index, question_number, question, variant):
        return self._create_answers_text(question_index, question_number, question, variant)

    def create_answers_essay(self, question_index, question_number, question, variant):
        return self._create_answers_text(question_index, question_number, question, variant)

    def _create_answers_text(self, question_index, question_number, question, variant):
        document = question.answers_documents[0]

        return {
            'raw_solutions': question.answers,
            'solution': self.clean_solution_content(document),
            'dirty_solution': document.to_format(self.format),
        }



import datetime
import json
import logging
import os
import random

import json5

import quizgen.common
import quizgen.group
import quizgen.parser
import quizgen.uploader.canvas
import quizgen.util.file
import quizgen.util.git

class Quiz(object):
    def __init__(self,
            title = '',
            course_title = '', term_title = '',
            description = '', date = '',
            time_limit_mins = None, shuffle_answers = True,
            groups = [],
            base_dir = '.',
            version = None,
            canvas = {},
            **kwargs):
        self.title = title
        self.course_title = course_title
        self.term_title = term_title
        self.date = date

        self.description = description
        self.description_document = None

        self.time_limit_mins = time_limit_mins
        self.shuffle_answers = shuffle_answers

        self.groups = groups

        self.version = version
        self.base_dir = base_dir

        self.canvas = canvas.copy()

        try:
            self.validate(kwargs)
        except Exception as ex:
            raise quizgen.common.QuizValidationError("Error while validating quiz '%s'." % self.title) from ex

    def validate(self, kwargs):
        if ((self.title is None) or (self.title == "")):
            raise quizgen.common.QuizValidationError("Title cannot be empty.")

        if ((self.description is None) or (self.description == "")):
            raise quizgen.common.QuizValidationError("Description cannot be empty.")
        self.description_document = quizgen.parser.parse_text(self.description,
                base_dir = self.base_dir)

        if (self.version is None):
            self.version = quizgen.util.git.get_version(self.base_dir, throw = False)
            if (self.version == quizgen.util.git.UNKNOWN_VERSION):
                logging.warning("Could not get a version for the quiz (is it in a git repo?).")

        self.canvas = quizgen.uploader.canvas.validate_options(self.canvas)

        self._validate_time_limit()

        if (self.date == ''):
            self.date = datetime.date.today()
        else:
            self.date = datetime.date.fromisoformat(self.date)

        for key in kwargs:
            logging.warning("Unknown quiz option: '%s'." % (key))

    def _validate_time_limit(self):
        if (self.time_limit_mins is None):
            return

        if (not isinstance(self.time_limit_mins, (str, int))):
            raise quizgen.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        try:
            self.time_limit_mins = int(self.time_limit_mins)
        except:
            raise quizgen.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        if (self.time_limit_mins < 0):
            raise quizgen.common.QuizValidationError("Time limit must be a positive int, found '%s'." % (str(self.time_limit_mins)))

        if (self.time_limit_mins == 0):
            self.time_limit_mins = None

    def to_dict(self, include_docs = True, flatten_groups = False):
        value = self.__dict__.copy()

        if ('date' in value):
            value['date'] = value['date'].isoformat()

        value['groups'] = [group.to_dict(include_docs = include_docs) for group in self.groups]

        if (include_docs):
            value['description_document'] = self.description_document.to_pod()
        else:
            del value['description_document']

        return value

    @staticmethod
    def from_path(path, flatten_groups = False):
        path = os.path.abspath(path)

        with open(path, 'r') as file:
            quiz_info = json5.load(file)

        # Check for a description file.
        description_filename = os.path.splitext(os.path.basename(path))[0]
        description_path = os.path.join(os.path.dirname(path), description_filename + '.md')
        if (os.path.exists(description_path)):
            quiz_info['description'] = quizgen.util.file.read(description_path)
            logging.debug("Loading quiz description from '%s'.", description_path)

        base_dir = os.path.dirname(path)

        return Quiz.from_dict(quiz_info, base_dir, flatten_groups = flatten_groups)

    @staticmethod
    def from_dict(quiz_info, base_dir = None, flatten_groups = False):
        groups = [quizgen.group.Group.from_dict(group_info, base_dir) for group_info in quiz_info.get('groups', [])]

        if (flatten_groups):
            new_groups = []

            for old_group in groups:
                for i in range(len(old_group.questions)):
                    info = {
                        'name': old_group.name,
                        'pick_count': 1,
                        'points': old_group.points,
                        'questions': [old_group.questions[i]],
                    }

                    new_groups.append(quizgen.group.Group(**info))

            groups = new_groups

        quiz_info['groups'] = groups

        if (base_dir is not None):
            quiz_info['base_dir'] = base_dir
        elif ('base_dir' not in quiz_info):
            quiz_info['base_dir'] = '.'

        return Quiz(**quiz_info)

    def to_json(self, indent = 4, include_docs = True):
        return json.dumps(self.to_dict(include_docs = include_docs), indent = indent)

    def num_questions(self):
        count = 0

        for group in self.groups:
            count += group.pick_count

        return count

    def create_variant(self, identifier = None, seed = None, all_questions = False):
        if (seed is None):
            seed = random.randint(0, 2**64)

        logging.debug("Creating variant with seed %s.", str(seed))
        rng = random.Random(seed)

        questions = []
        for group in self.groups:
            new_questions = None

            if (all_questions):
                new_questions = group.copy_questions()
            else:
                new_questions = group.choose_questions(rng)

            if (len(new_questions) > 1):
                for i in range(len(new_questions)):
                    new_questions[i].base_name = "%s - %d" % (group.name, i + 1)

            questions += new_questions

        if (self.shuffle_answers):
            for question in questions:
                question.shuffle(rng)

        title = self.title
        version = self.version

        if (identifier is not None):
            title = "%s - %s" % (title, identifier)
            version = "%s, Variant: %s" % (version, identifier)

        return quizgen.variant.Variant(
            title = title,
            course_title = self.course_title, term_title = self.term_title,
            description = self.description, description_document = self.description_document,
            date = self.date,
            questions = questions,
            version = version, seed = seed)



import logging
import os
import shutil
import subprocess

def is_available():
    if (shutil.which('pdflatex') is None):
        logging.warning("Could not find `pdxlatex`, cannot compile PDFs")
        return False

    return True

def compile(path):
    result = subprocess.run(['pdflatex', '-interaction=nonstopmode', os.path.basename(path)], cwd = os.path.dirname(path),
        capture_output = True)

    if (result.returncode != 0):
        raise ValueError("pdflatex did not exit cleanly. Stdout: '%s', Stderr: '%s'" % (result.stdout, result.stderr))



FORMAT_HTML = 'html'
FORMAT_JSON = 'json'
FORMAT_MD = 'md'
FORMAT_TEX = 'tex'
FORMAT_TEXT = 'text'

# Output formats.
DOC_FORMATS = [FORMAT_HTML, FORMAT_JSON, FORMAT_MD, FORMAT_TEX, FORMAT_TEXT]

# Formats that the parser understands basic conversions to.
PARSER_FORMATS = [FORMAT_HTML, FORMAT_JSON, FORMAT_MD, FORMAT_TEX, FORMAT_TEXT]

QUESTION_TYPE_ESSAY = 'essay'
QUESTION_TYPE_FIMB = 'fill_in_multiple_blanks'
QUESTION_TYPE_FITB = 'fill_in_the_blank'
QUESTION_TYPE_MATCHING = 'matching'
QUESTION_TYPE_MA = 'multiple_answers'
QUESTION_TYPE_MCQ = 'multiple_choice'
QUESTION_TYPE_MDD = 'multiple_dropdowns'
QUESTION_TYPE_NUMERICAL = 'numerical'
QUESTION_TYPE_SA = 'short_answer'
QUESTION_TYPE_TEXT_ONLY = 'text_only'
QUESTION_TYPE_TF = 'true_false'

# Supported question types.
QUESTION_TYPES = [
    QUESTION_TYPE_ESSAY,
    QUESTION_TYPE_FIMB,
    QUESTION_TYPE_FITB,
    QUESTION_TYPE_MATCHING,
    QUESTION_TYPE_MA,
    QUESTION_TYPE_MCQ,
    QUESTION_TYPE_MDD,
    QUESTION_TYPE_NUMERICAL,
    QUESTION_TYPE_SA,
    QUESTION_TYPE_TEXT_ONLY,
    QUESTION_TYPE_TF,
]

NUMERICAL_ANSWER_TYPE_EXACT = 'exact'
NUMERICAL_ANSWER_TYPE_RANGE = 'range'
NUMERICAL_ANSWER_TYPE_PRECISION = 'precision'



"""
Create (and possibly upload) a quiz in the GradeScope format.
"""

import argparse
import datetime
import json
import logging
import os
import random
import string
import sys

import quizgen.converter.textemplate
import quizgen.latex
import quizgen.log
import quizgen.uploader.gradescope
import quizgen.util.file
import quizgen.quiz

OPTIONS_FILENAME = 'options.json'

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    if (args.upload):
        if ((args.course_id is None) or (args.user is None) or (args.password is None)):
            raise ValueError("If --upload is provided, then --course, --user, and --pass are all required.")

    if ((args.variants < 1) or (args.variants >= len(string.ascii_uppercase))):
        raise ValueError("Number of variants must be in [1, %d), found %d." % (len(string.ascii_uppercase), args.variants))

    quiz = quizgen.quiz.Quiz.from_path(args.path)

    out_dir = os.path.join(args.out_dir, quiz.title)
    os.makedirs(out_dir, exist_ok = True)

    logging.info("Writing generated output to '%s'.", out_dir)

    seed = args.seed
    if (seed is None):
        seed = random.randint(0, 2**64)

    options = {
        'create_time': datetime.datetime.now().isoformat(),
        'seed': seed,
        'out_dir': out_dir,
        'quiz': {
            'path': args.path,
            'title': quiz.title,
            'version': quiz.version,
        },
        'gradescope': {
            'upload': args.upload,
            'course': args.course_id,
            'rubric': args.rubric,
            'user': args.user,
        },
        'variants': {
            'count': args.variants,
            'titles': [],
            'ids': [],
        },
    }

    logging.info("Using seed %d.", seed)
    rng = random.Random(seed)

    gradescope_ids = []

    for i in range(args.variants):
        variant_id = None
        if (args.variants > 1):
            variant_id = string.ascii_uppercase[i]

        variant = quiz.create_variant(identifier = variant_id, seed = rng.randint(0, 2**64))
        out_path = os.path.join(out_dir, "%s.json" % (variant.title))
        quizgen.util.file.write(out_path, variant.to_json(include_docs = False))

        options['variants']['titles'].append(variant.title)

        if (args.upload):
            uploader = quizgen.uploader.gradescope.GradeScopeUploader(args.course_id, args.user, args.password,
                    force = args.force, rubric = args.rubric)
            gradescope_id, created = uploader.upload_quiz(variant, base_dir = out_dir)

            if (created):
                gradescope_ids.append(gradescope_id)

            options['variants']['ids'].append(gradescope_id)
        else:
            _make_pdf(variant, out_dir, False)
            options['variants']['ids'].append(None)

        title = variant.title

        # Always create an answer key.
        try:
            variant.title = "%s -- Answer Key" % (variant.title)
            _make_pdf(variant, out_dir, True)
        except Exception as ex:
            logging.warning("Failed to generate answer key for '%s'.", title)

        logging.info("Completed variant: '%s'.", title)

    # If there are multiple variants and all variants were created (non were skipped).
    if ((args.variants > 1) and (len(gradescope_ids) == args.variants)):
        uploader = quizgen.uploader.gradescope.GradeScopeUploader(args.course_id, args.user, args.password)
        uploader.create_assignment_group(quiz.title, gradescope_ids)
        logging.info("Created GradeScope Assignment Group: '%s'.", quiz.title)

    path = os.path.join(out_dir, OPTIONS_FILENAME)
    with open(path, 'w') as file:
        json.dump(options, file, indent = 4)

    return 0

def _make_pdf(variant, out_dir, is_key):
    converter = quizgen.converter.textemplate.TexTemplateConverter(answer_key = is_key)
    content = converter.convert_variant(variant)

    out_path = os.path.join(out_dir, "%s.tex" % (variant.title))
    quizgen.util.file.write(out_path, content)

    # Need to compile twice to get positioning.
    quizgen.latex.compile(out_path)
    quizgen.latex.compile(out_path)

def _get_parser():
    parser = argparse.ArgumentParser(description =
        "Parse a quiz and potentially upload it to GradeScope.")

    parser.add_argument('path',
        type = str,
        help = 'The path to a quiz json file.')

    parser.add_argument('--variants', dest = 'variants',
        action = 'store', type = int, default = 1,
        help = 'The number of quiz variants to create (default: %(default)s).')

    parser.add_argument('--outdir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = 'The directory to put the quiz creation output (which will be another directory) (default: %(default)s).')

    parser.add_argument('--upload', dest = 'upload',
        action = 'store_true', default = False,
        help = 'Upload the quiz(zes) to GradeScope, requires --course, --user, and --pass (default: %(default)s).')

    parser.add_argument('--rubric', dest = 'rubric',
        action = 'store_true', default = False,
        help = 'Create an initial rubric for this assignment in GradeScope (default: %(default)s).')

    parser.add_argument('--course', dest = 'course_id',
        action = 'store', type = str,
        help = 'Course ID to upload the quiz under.')

    parser.add_argument('--user', dest = 'user',
        action = 'store', type = str,
        help = 'The user to authenticate as.')

    parser.add_argument('--pass', dest = 'password',
        action = 'store', type = str,
        help = 'The password to use.')

    parser.add_argument('--force', dest = 'force',
        action = 'store_true', default = False,
        help = 'Override (delete) any exiting quiz with the same name.')

    parser.add_argument('--seed', dest = 'seed',
        action = 'store', type = int, default = None,
        help = 'The random seed to use (defaults to a random seed).')

    quizgen.log.set_cli_args(parser)

    return parser

def main():
    args = _get_parser().parse_args()
    quizgen.log.init_from_args(args)
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import abc
import base64
import copy
import html
import json
import os
import re

import lark
import lark.visitors

import quizgen.katex
import quizgen.util.file

ENCODING = 'utf-8'

GRAMMAR = r'''
    document: [ block ( NEWLINE+ block )* NEWLINE* ]

    block: ( ( code_block | equation_block | table_block | list_block | text_line ) NEWLINE )+

    code_block: "```" NEWLINE? code_block_internal "```"
    ?code_block_internal: /.+?(?=```)/s
    CODE_BLOCK_TERMINAL: "```"

    equation_block: "$$" NEWLINE? equation_block_internal "$$"
    ?equation_block_internal: /.+?(?=\$\$)/s

    table_block: ( ( table_head | table_row | table_sep ) NEWLINE )+
    table_sep: /\|-{3}[-\s]*\|.*/
    table_row: "|" table_cell+
    table_head: "|-" table_cell+
    table_cell: text_line "|"

    list_block: ( LIST_ITEM_START text_line NEWLINE )+
    LIST_ITEM_START: /\s*-/

    text_line: ( text_line_internal )+
    ?text_line_internal: inline_comment
                       | inline_code
                       | inline_equation
                       | inline_italics
                       | inline_bold
                       | inline_link
                       | inline_image
                       | inline_linebreak
                       | inline_answer_reference
                       | inline_text

    inline_comment: /\s*\/\// INLINE_COMMENT
    inline_link: INLINE_LINK_TEXT INLINE_LINK_LINK
    inline_image: "!" INLINE_LINK_TEXT INLINE_LINK_LINK
    inline_code: INLINE_CODE
    inline_equation: INLINE_EQUATION
    inline_italics: INLINE_ITALICS
    inline_bold: INLINE_BOLD
    inline_linebreak: "\\n"
    inline_answer_reference: "[[" REFERENCE_WORD "]]"
    inline_text: ( ESC_CHAR | NON_ESC_TEXT )+

    _ESCAPE_INTERNAL: /.+?/ /(?<!\\)(\\\\)*?/

    INLINE_COMMENT: /.*?\n/s
    INLINE_CODE: "`" _ESCAPE_INTERNAL "`"
    INLINE_EQUATION: "$" _ESCAPE_INTERNAL "$"
    INLINE_ITALICS: "*" _ESCAPE_INTERNAL "*"
    INLINE_BOLD: "**" _ESCAPE_INTERNAL "**"
    INLINE_LINK_TEXT: "[" _ESCAPE_INTERNAL "]"
    INLINE_LINK_LINK: "(" _ESCAPE_INTERNAL ")"

    REFERENCE_WORD: /[a-zA-Z][a-zA-Z0-9_]*/

    NON_ESC_TEXT: NON_ESC_CHAR+
    NON_ESC_CHAR: /[^\n\\`|\*\$\-\[!\/]/x
    ESC_CHAR: "\\\\"
            | "\\-"
            | "\\*"
            | "\\|"
            | "\\$"
            | "\\["
            | "\\!"
            | "\\`"
            | "\\/"

    NEWLINE: /\n/
'''

TEX_REPLACEMENTS = {
    '\\': '\\textbackslash{}',
    '|': '\\textbar{}',
    '$': '\\$',
    '#': '\\#',
    '%': '\\%',
    '_': '\\_',
    'π': '$\\pi$',
    '`': '\\`{}',
    '{': '\\{',
    '}': '\\}',
}

HTML_TABLE_STYLE = [
    'border-collapse: collapse',
    'text-align: center',
    'margin-bottom: 1em',
]

VERB_CHARACTERS = ['|', '!', '@', '#', '$', '^', '&', '-', '_', '=', '+']

class DocTransformer(lark.Transformer):
    def document(self, blocks):
        return DocumentNode(blocks)

    def block(self, nodes):
        return BlockNode(nodes)

    def code_block(self, text):
        # Trim any newlines.
        text = text[0].strip("\n")
        return CodeNode(text, inline = False)

    def equation_block(self, text):
        # Trim any newlines.
        text = text[0].strip()
        return EquationNode(text, inline = False)

    def text_line(self, nodes):
        return TextNode(nodes)

    def inline_comment(self, text):
        return CommentNode(str(text[1]).strip())

    def inline_text(self, text):
        return NormalTextNode(''.join(text))

    def inline_linebreak(self, _):
        return LinebreakNode()

    def inline_answer_reference(self, text):
        return AnswerReferenceNode(str(text[0]))

    def inline_italics(self, text):
        # Strip off the asterisks.
        text = str(text[0])[1:-1]
        return ItalicsNode(text)

    def inline_bold(self, text):
        # Strip off the asterisks.
        text = str(text[0])[2:-2]
        return BoldNode(text)

    def inline_code(self, text):
        # Strip off the backticks.
        text = str(text[0])[1:-1]

        # Replace any escaped backticks.
        text = text.replace(r'\`', '`')

        return CodeNode(text, inline = True)

    def inline_equation(self, text):
        # Strip off the dollar signs.
        text = str(text[0])[1:-1].strip()

        # Replace any escaped dollar signs.
        text = text.replace(r'\$', '$')

        return EquationNode(text, inline = True)

    def table_block(self, rows):
        return TableNode(rows)

    def table_row(self, cells):
        return TableRowNode(cells, head = False)

    def table_head(self, cells):
        return TableRowNode(cells, head = True)

    def table_sep(self, cells):
        return TableSepNode()

    def table_cell(self, cell):
        return cell[0].trim()

    def list_block(self, items):
        return ListNode([item.trim() for item in items])

    def inline_link(self, contents):
        # Remove the surrounding characters, strip it, and replace escaped end markers.
        text = str(contents[0])[1:-1].strip().replace(r'\]', ']')
        link = str(contents[1])[1:-1].strip().replace(r'\)', ')')

        return LinkNode(text, link)

    def inline_image(self, contents):
        text = str(contents[0])[1:-1].strip()
        link = str(contents[1])[1:-1].strip()

        return ImageNode(text, link)

    def NON_ESC_TEXT(self, text):
        return str(text)

    def ESC_CHAR(self, text):
        # Remove the backslash.
        return text[1:]

    def LIST_ITEM_START(self, token):
        return lark.visitors.Discard

    def NEWLINE(self, token):
        return lark.visitors.Discard

class ParseNode(abc.ABC):
    @abc.abstractmethod
    def to_markdown(self, **kwargs):
        pass

    @abc.abstractmethod
    def to_tex(self, **kwargs):
        pass

    @abc.abstractmethod
    def to_html(self, **kwargs):
        pass

    @abc.abstractmethod
    def to_pod(self, **kwargs):
        """
        Get a "Plain Old Data" representation of the node."
        This representation should be convertable to JSON."
        """

        pass

    def collect_file_paths(self, base_dir):
        return []

    def trim(self, left = True, right = True):
        pass

    def is_empty(self):
        return False

    def to_json(self, indent = 4, **kwargs):
        return json.dumps(self.to_pod(**kwargs), indent = indent)

    def to_format(self, format, **kwargs):
        if (format == quizgen.constants.FORMAT_HTML):
            return self.to_html(**kwargs)
        elif (format == quizgen.constants.FORMAT_JSON):
            return self.to_json(**kwargs)
        elif (format == quizgen.constants.FORMAT_MD):
            return self.to_markdown(**kwargs)
        elif (format == quizgen.constants.FORMAT_TEX):
            return self.to_tex(**kwargs)
        elif (format == quizgen.constants.FORMAT_TEXT):
            return self.to_text(**kwargs)
        else:
            raise ValueError(f"Unknown format '{format}'.")

class DocumentNode(ParseNode):
    def __init__(self, nodes):
        self._nodes = list(nodes)
        self._context = {}

    def set_context(self, key, value):
        self._context[key] = value

    def set_base_dir(self, base_dir):
        self.set_context("base_dir", base_dir)

    def to_markdown(self, **kwargs):
        context = copy.deepcopy(self._context)
        context.update(kwargs)

        return "\n\n".join([node.to_markdown(**context) for node in self._nodes])

    def to_text(self, **kwargs):
        context = copy.deepcopy(self._context)
        context.update(kwargs)

        return "\n\n".join([node.to_text(**context) for node in self._nodes])

    def to_tex(self, **kwargs):
        context = copy.deepcopy(self._context)
        context.update(kwargs)

        content = "\n\n".join([node.to_tex(**context) for node in self._nodes])
        return content

    def to_html(self, **kwargs):
        context = copy.deepcopy(self._context)
        context.update(kwargs)

        content = "\n\n".join([node.to_html(level = 1, **context) for node in self._nodes])
        content = f"<div class='document'>\n{content}\n</div>"

        return content

    def to_pod(self, include_metadata = True, **kwargs):
        data = {
            "type": "document",
            "nodes": [node.to_pod(include_metadata = True, **kwargs) for node in self._nodes],
        }

        if (include_metadata):
            data["context"] = self._context

        return data

    def collect_file_paths(self, base_dir):
        paths = []

        for node in self._nodes:
            paths += node.collect_file_paths(base_dir)

        return paths

class BlockNode(ParseNode):
    def __init__(self, nodes):
        self._nodes = list(nodes)

    def to_markdown(self, **kwargs):
        return "\n".join([node.to_markdown(**kwargs) for node in self._nodes])

    def to_text(self, **kwargs):
        return "\n".join([node.to_text(**kwargs) for node in self._nodes])

    def to_tex(self, **kwargs):
        return "\n".join([node.to_tex(**kwargs) for node in self._nodes])

    def to_html(self, **kwargs):
        content = "\n".join([node.to_html(**kwargs) for node in self._nodes])
        return f"<div class='block' style='margin-bottom: 1em;'>\n{content}\n</div>"

    def to_pod(self, **kwargs):
        return {
            "type": "block",
            "nodes": [node.to_pod(**kwargs) for node in self._nodes],
        }

    def collect_file_paths(self, base_dir):
        paths = []

        for node in self._nodes:
            paths += node.collect_file_paths(base_dir)

        return paths

class LinkNode(ParseNode):
    def __init__(self, text, link):
        self._text = text
        self._link = link

    def to_markdown(self, **kwargs):
        return f"[{self._text}]({self._link})"

    def to_text(self, **kwargs):
        return f"{self._text} ({self._link})"

    def to_tex(self, **kwargs):
        if (self._text == ""):
            return rf"\url{{{self._link}}}"

        return rf"\href{{{self._link}}}{{{self._text}}}"

    def to_html(self, **kwargs):
        text = self._text
        if (text == ''):
            text = self._link

        return f"<a href='{self._link}'>{text}</a>"

    def to_pod(self, **kwargs):
        return {
            "type": "link",
            "text": self._text,
            "link": self._link,
        }

class ImageNode(ParseNode):
    def __init__(self, text, link):
        self._text = text
        self._link = link

    def to_markdown(self, **kwargs):
        return f"![{self._text}]({self._link})"

    def to_text(self, **kwargs):
        return f"{self._text} ({self._link})"

    def to_tex(self, base_dir = '.', **kwargs):
        path = os.path.join(base_dir, self._link)
        return rf"\includegraphics[width=0.5\textwidth]{{{path}}}"

    def to_html(self, base_dir = '.', canvas_instance = None, **kwargs):
        if (re.match(r'^http(s)?://', self._link)):
            return f"<img src='{self._link}' alt='{self._text}' />"

        path = os.path.realpath(os.path.join(base_dir, self._link))

        if (canvas_instance is None):
            # If we are not uploading to canvas, do a base64 encode of the image.
            mime, content = encode_image(path)
            src = f"data:{mime};base64,{content}"
            return f"<img src='{src}' alt='{self._text}' />"

        # Canvas requires uploading the image, which should have been done via quizgen.converter.canvas.upload_canvas_files().

        file_id = canvas_instance.context.get('file_ids', {}).get(path)
        if (file_id is None):
            raise ValueError(f"Could not get canvas context file id of image '{path}'.")

        src = f"{canvas_instance.base_url}/courses/{canvas_instance.course_id}/files/{file_id}/preview"
        return f"<img src='{src}' alt='{self._text}' />"

    def collect_file_paths(self, base_dir):
        if (re.match(r'^http(s)?://', self._link)):
            return []

        path = os.path.realpath(os.path.join(base_dir, self._link))
        return [path]

    def to_pod(self, **kwargs):
        return {
            "type": "image",
            "text": self._text,
            "link": self._link,
        }

class TableNode(ParseNode):
    def __init__(self, rows):
        self._rows = list(rows)

        self._width = 0
        for row in self._rows:
            self._width = max(self._width, len(row))

    def to_markdown(self, **kwargs):
        return "\n".join([row.to_markdown(width = self._width, **kwargs) for row in self._rows]) + "\n"

    def to_text(self, **kwargs):
        return "\n".join([row.to_text(width = self._width, **kwargs) for row in self._rows]) + "\n"

    def to_tex(self, **kwargs):
        column_spec = "c" * self._width

        lines = [
            r'\begin{center}',
            r'    \begin{tabular}{ ' + column_spec + ' }',
        ]

        for row in self._rows:
            row = row.to_tex(width = self._width, **kwargs)
            lines.append(f"        {row}")

        lines += [
            r'    \end{tabular}',
            r'\end{center}',
            '',
        ]

        return "\n".join(lines)

    def to_html(self, **kwargs):
        table_style = ' ; '.join(HTML_TABLE_STYLE)

        lines = [
            f'<table style="{table_style}">',
        ]

        next_cell_style = None
        for row in self._rows:
            if (isinstance(row, TableSepNode)):
                next_cell_style = "border-top: 2px solid black;"
            else:
                lines.append(row.to_html(cell_style = next_cell_style, **kwargs))
                next_cell_style = None

        lines += [
            '</table>'
            '',
        ]

        return "\n".join(lines)

    def to_pod(self, **kwargs):
        return {
            "type": "table",
            "rows": [row.to_pod(**kwargs) for row in self._rows],
        }

    def collect_file_paths(self, base_dir):
        paths = []

        for row in self._rows:
            paths += row.collect_file_paths(base_dir)

        return paths

class TableRowNode(ParseNode):
    def __init__(self, cells, head = False):
        self._cells = list(cells)
        self._head = head

    def to_markdown(self, **kwargs):
        return "| " + " | ".join([cell.to_markdown(**kwargs) for cell in self._cells]) + " |"

    def to_text(self, **kwargs):
        return "| " + " | ".join([cell.to_text(**kwargs) for cell in self._cells]) + " |"

    def to_tex(self, **kwargs):
        return " & ".join([cell.to_tex(**kwargs) for cell in self._cells]) + r' \\'

    def to_html(self, cell_style = None, **kwargs):
        tag = 'td'
        if (self._head):
            tag = 'th'

        cell_inline_style = "padding-left: 0.50em ; padding-right: 0.50em ; padding-bottom: 0.25em "
        if (cell_style is not None):
            cell_inline_style += (' ; ' + cell_style)

        lines = ['<tr>']

        for cell in self._cells:
            content = cell.to_html(**kwargs)
            content = f"<{tag} style='{cell_inline_style}' >{content}</{tag}>"
            lines.append(content)

        lines += ['</tr>']

        return "\n".join(lines)

    def to_pod(self, **kwargs):
        return {
            "type": "table-row",
            "head": self._head,
            "cells": [cell.to_pod(**kwargs) for cell in self._cells],
        }

    def collect_file_paths(self, base_dir):
        paths = []

        for cell in self._cells:
            paths += cell.collect_file_paths(base_dir)

        return paths

    def __len__(self):
        return len(self._cells)

class TableSepNode(ParseNode):
    def __init__(self):
        pass

    def to_markdown(self, width = 1, **kwargs):
        return "|" + ("---|" * width)

    def to_text(self, width = 1, **kwargs):
        return "|" + ("---|" * width)

    def to_tex(self, **kwargs):
        return r'\hline \\'

    def to_html(self, **kwargs):
        raise RuntimeError("to_html() should never be called on a table separator (row should handle it).")

    def to_pod(self, **kwargs):
        return {
            "type": "table-sep",
        }

    def __len__(self):
        return 1

class ListNode(ParseNode):
    def __init__(self, items):
        self._items = list(items)

    def to_markdown(self, **kwargs):
        return "\n".join([" - " + item.to_markdown(**kwargs) for item in self._items]) + "\n"

    def to_text(self, **kwargs):
        return "\n".join([" - " + item.to_text(**kwargs) for item in self._items]) + "\n"

    def to_tex(self, **kwargs):
        lines = [
            r'\begin{itemize}',
        ]

        for item in self._items:
            text = item.to_tex(**kwargs)
            lines.append(f"    \item {text}")

        lines += [
            r'\end{itemize}',
            '',
        ]

        return "\n".join(lines)

    def to_html(self, **kwargs):
        lines = [
            '<ul style="margin-bottom: 0; margin-top: 0;">',
        ]

        for item in self._items:
            text = item.to_html(**kwargs)
            lines.append(f'<li>{text}</li>')

        lines += [
            '</ul>'
            '',
        ]

        return "\n".join(lines)

    def to_pod(self, **kwargs):
        return {
            "type": "list",
            "items": [item.to_pod(**kwargs) for item in self._items],
        }

    def collect_file_paths(self, base_dir):
        paths = []

        for item in self._items:
            paths += item.collect_file_paths(base_dir)

        return paths

class TextNode(ParseNode):
    def __init__(self, nodes):
        self._nodes = list(nodes)

    def to_markdown(self, **kwargs):
        return "".join([node.to_markdown(**kwargs) for node in self._nodes])

    def to_text(self, **kwargs):
        return "".join([node.to_text(**kwargs) for node in self._nodes])

    def to_tex(self, **kwargs):
        return "".join([node.to_tex(**kwargs) for node in self._nodes])

    def to_html(self, **kwargs):
        return "".join([node.to_html(**kwargs) for node in self._nodes])

    def to_pod(self, **kwargs):
        return {
            "type": "text",
            "nodes": [node.to_pod(**kwargs) for node in self._nodes],
        }

    def trim(self, left = True, right = True):
        """
        Trim the whitespace off the edges of this node.
        """

        if (left):
            self._nodes[0].trim(right = False)

        if (right):
            self._nodes[-1].trim(left = False)

        if (self._nodes[0].is_empty()):
            self._nodes.pop(0)

        if ((len(self._nodes) > 0) and (self._nodes[-1].is_empty())):
            self._nodes.pop()

        return self

    def collect_file_paths(self, base_dir):
        paths = []

        for node in self._nodes:
            paths += node.collect_file_paths(base_dir)

        return paths

class LinebreakNode(ParseNode):
    def __init__(self):
        pass

    def to_markdown(self, **kwargs):
        return "  \n"

    def to_text(self, **kwargs):
        return "\n\n"

    def to_tex(self, **kwargs):
        return ' \\newline\n'

    def to_html(self, **kwargs):
        return "<br />"

    def to_pod(self, **kwargs):
        return {
            "type": "linebreak",
        }

class BaseTextNode(ParseNode):
    def __init__(self, text, type):
        self._text = text
        self._type = type

    def to_pod(self, **kwargs):
        return {
            "type": self._type,
            "text": self._text,
        }

    def trim(self, left = True, right = True):
        if (left):
            self._text = self._text.lstrip()

        if (right):
            self._text = self._text.rstrip()

        return self

    def is_empty(self):
        return (len(self._text) == 0)

class AnswerReferenceNode(BaseTextNode):
    def __init__(self, text):
        super().__init__(text, 'answer-reference')

    def to_markdown(self, **kwargs):
        return f"[[{self._text}]]"

    def to_text(self, **kwargs):
        return f"[{self._text}]"

    def to_tex(self, **kwargs):
        # TODO(eriq): We do not have a convention for this.
        text = tex_escape(self._text)
        return rf"\textsc{{<{text}>}}"

    def to_html(self, **kwargs):
        text = html.escape(self._text)
        return f"<span>[{text}]</span>"

class CommentNode(BaseTextNode):
    def __init__(self, text):
        super().__init__(text, 'comment')

    def to_markdown(self, display_comments = False, **kwargs):
        if (not display_comments):
            return ""

        return f"<!--- {self._text} -->"

    def to_text(self, display_comments = False, **kwargs):
        if (not display_comments):
            return ""

        return f"# {self._text}"

    def to_tex(self, display_comments = False, **kwargs):
        if (not display_comments):
            return ""

        return f"% {self._text}"

    def to_html(self, display_comments = False, **kwargs):
        if (not display_comments):
            return ""

        text = html.escape(self._text)
        return f"<!--- {text} -->"

class NormalTextNode(BaseTextNode):
    def __init__(self, text):
        super().__init__(text, 'normal_text')

    def to_markdown(self, **kwargs):
        return self._text

    def to_text(self, **kwargs):
        return self._text

    def to_tex(self, **kwargs):
        return tex_escape(self._text)

    def to_html(self, **kwargs):
        text = html.escape(self._text)
        return f"<span>{text}</span>"

class ItalicsNode(BaseTextNode):
    def __init__(self, text):
        super().__init__(text, 'italics_text')

    def to_markdown(self, **kwargs):
        return f"*{self._text}*"

    def to_text(self, **kwargs):
        return self._text

    def to_tex(self, **kwargs):
        text = tex_escape(self._text)
        return rf"\textit{{{text}}}"

    def to_html(self, **kwargs):
        text = html.escape(self._text)
        return f"<span><emph>{text}</emph></span>"

class BoldNode(BaseTextNode):
    def __init__(self, text):
        super().__init__(text, 'bold_text')

    def to_markdown(self, **kwargs):
        return f"**{self._text}**"

    def to_text(self, **kwargs):
        return self._text

    def to_tex(self, **kwargs):
        text = tex_escape(self._text)
        return rf"\textbf{{{text}}}"

    def to_html(self, **kwargs):
        text = html.escape(self._text)
        return f"<span><strong>{text}</strong></span>"

class CodeNode(BaseTextNode):
    def __init__(self, text, inline = False):
        super().__init__(text, 'code')
        self._inline = inline

    def to_markdown(self, **kwargs):
        if (self._inline):
            return f"`{self._text}`"

        return f"```\n{self._text}\n```"

    def to_text(self, bare = False, **kwargs):
        if (bare and self._inline):
            return f"{self._text}"
        elif (bare and (not self._inline)):
            return f"\n{self._text}\n"
        elif ((not bare) and self._inline):
            return f"`{self._text}`"
        else:
            return f"```\n{self._text}\n```"

    def to_tex(self, **kwargs):
        if (not self._inline):
            return f"\\begin{{lstlisting}}\n{self._text}\n\\end{{lstlisting}}"

        delim = None
        for char in VERB_CHARACTERS:
            if (char not in self._text):
                delim = char
                break

        if (delim is None):
            raise ValueError("Could not find a delimiter to use with tex's `\verb'.")

        return r"\verb%s%s%s" % (delim, self._text, delim)

    def to_html(self, **kwargs):
        content = f'<code>{self._text}</code>'

        if (not self._inline):
            content = f"<pre style='margin: 1em'>{content}</pre>"

        return content

    def to_pod(self, **kwargs):
        value = super().to_pod(**kwargs)
        value["inline"] = self._inline

        return value

class EquationNode(BaseTextNode):
    katex_available = None

    def __init__(self, text, inline = False):
        super().__init__(text, 'equation')
        self._inline = inline

    def to_markdown(self, **kwargs):
        if (self._inline):
            return f"$ {self._text} $"

        return f"$$\n{self._text}\n$$"

    def to_text(self, bare = False, **kwargs):
        if (bare and self._inline):
            return f"{self._text}"
        elif (bare and (not self._inline)):
            return f"\n{self._text}\n"
        elif ((not bare) and self._inline):
            return f"$ {self._text} $"
        else:
            return f"$$\n{self._text}\n$$"

    def to_tex(self, **kwargs):
        if (self._inline):
            return f"$ {self._text} $"

        return f"$$\n{self._text}\n$$"

    def to_html(self, **kwargs):
        if (EquationNode.katex_available is None):
            EquationNode.katex_available = quizgen.katex.is_available()

        content = f"<code>{self._text}</code>"
        if (EquationNode.katex_available):
            content = quizgen.katex.to_html(self._text)

        element = 'p'
        if (self._inline):
            element = 'span'

        return f"<{element}>{content}</{element}>"

    def to_pod(self, **kwargs):
        value = super().to_pod(**kwargs)
        value["inline"] = self._inline

        return value

def encode_image(path):
    ext = os.path.splitext(path)[-1].lower()
    mime = f"image/{ext}"

    with open(path, 'rb') as file:
        data = file.read()

    content = base64.standard_b64encode(data)
    return mime, content.decode(ENCODING)

def tex_escape(text):
    """
    Prepare normal text for tex.
    """

    for key, value in TEX_REPLACEMENTS.items():
        text = text.replace(key, value)

    return text

def clean_text(text):
    # Remove carriage returns.
    text = text.replace("\r", '')

    # Trim whitespace.
    text = text.strip();

    # Replace the final newline and add one additional one (for tables).
    text += "\n\n"

    return text

def parse_text(text, base_dir = '.'):
    # Special case for empty documents.
    if (text.strip() == ''):
        return DocumentNode([])

    text = clean_text(text)

    parser = lark.Lark(GRAMMAR, start = 'document')
    ast = parser.parse(text)

    document = DocTransformer().transform(ast)
    document.set_base_dir(base_dir)

    return document

def parse_file(path):
    if (not os.path.isfile(path)):
        raise ValueError(f"Path to parse ('{path}') is not a file.")

    text = quizgen.util.file.read(path)
    base_dir = os.path.dirname(path)

    return parse_text(text, base_dir = base_dir)



"""
Upload a gradescope quiz that has already been created via create-gradescope-quiz.
"""

import argparse
import logging
import os
import string
import sys

import quizgen.converter.textemplate
import quizgen.latex
import quizgen.log
import quizgen.uploader.gradescope
import quizgen.util.file
import quizgen.variant

def run(args):
    if (not os.path.exists(args.path)):
        raise ValueError(f"Provided path '{args.path}' does not exist.")

    if (not os.path.isfile(args.path)):
        raise ValueError(f"Provided path '{args.path}' is not a file.")

    variant_path = os.path.abspath(args.path)
    base_dir = os.path.dirname(variant_path)

    variant = quizgen.variant.Variant.from_path(variant_path)

    logging.info("Writing generated output to '%s'.", base_dir)

    uploader = quizgen.uploader.gradescope.GradeScopeUploader(args.course_id, args.user, args.password,
            force = args.force, rubric = args.rubric,
            skip_tex_write = args.skip_tex_write, skip_compile = args.skip_compile)
    gradescope_id = uploader.upload_quiz(variant, base_dir = base_dir)

    title = variant.title

    try:
        variant.title = "%s -- Answer Key" % (variant.title)
        _make_pdf(variant, base_dir, True, args.skip_tex_write, args.skip_compile)
    except Exception as ex:
        logging.warning("Failed to generate answer key for '%s'.", title)

    logging.info("Completed variant: '%s'.", title)

    return 0

def _make_pdf(variant, base_dir, is_key, skip_tex_write, skip_compile):
    converter = quizgen.converter.textemplate.TexTemplateConverter(answer_key = is_key)
    content = converter.convert_variant(variant)

    out_path = os.path.join(base_dir, "%s.tex" % (variant.title))

    if (not skip_tex_write):
        quizgen.util.file.write(out_path, content)

    if (not skip_compile):
        # Need to compile twice to get positioning.
        quizgen.latex.compile(out_path)
        quizgen.latex.compile(out_path)

def _get_parser():
    parser = argparse.ArgumentParser(description =
        "Parse a single variant output generated by create-gradescope-quiz and upload the quiz to GradeScope."
        + " Most users should be using create-gradescope-quiz instead.")

    parser.add_argument('path',
        type = str,
        help = 'The path to a variant JSON file. It will be assumed that all the necessary supporting files are adjacent to this.')

    parser.add_argument('--rubric', dest = 'rubric',
        action = 'store_true', default = False,
        help = 'Create an initial rubric for this assignment in GradeScope (default: %(default)s).')

    parser.add_argument('--course', dest = 'course_id',
        action = 'store', type = str, required = True,
        help = 'Course ID to upload the quiz under.')

    parser.add_argument('--user', dest = 'user',
        action = 'store', type = str, required = True,
        help = 'The user to authenticate as.')

    parser.add_argument('--pass', dest = 'password',
        action = 'store', type = str, required = True,
        help = 'The password to use.')

    parser.add_argument('--force', dest = 'force',
        action = 'store_true', default = False,
        help = 'Override (delete) any exiting quiz with the same name.')

    parser.add_argument('--skip-tex', dest = 'skip_tex_write',
        action = 'store_true', default = False,
        help = 'Skip creating the tex file for this variant (assumes the tex file already exists).')

    parser.add_argument('--skip-compile', dest = 'skip_compile',
        action = 'store_true', default = False,
        help = 'Skip compiling the tex file for this variant (assumes the pdf is already compiled).')

    quizgen.log.set_cli_args(parser)

    return parser

def main():
    args = _get_parser().parse_args()
    quizgen.log.init_from_args(args)
    return run(args)

if (__name__ == '__main__'):
    sys.exit(main())



import quizgen.constants
import quizgen.converter.htmltemplate
import quizgen.converter.json
import quizgen.converter.textemplate
import quizgen.variant

SUPPORTED_FORMATS = [
    quizgen.constants.FORMAT_HTML,
    quizgen.constants.FORMAT_JSON,
    quizgen.constants.FORMAT_TEX,
]

def get_converter_class(format = quizgen.constants.FORMAT_JSON):
    if (format == quizgen.constants.FORMAT_JSON):
        return quizgen.converter.json.JSONConverter
    elif (format == quizgen.constants.FORMAT_HTML):
        return quizgen.converter.htmltemplate.HTMLTemplateConverter
    elif (format == quizgen.constants.FORMAT_TEX):
        return quizgen.converter.textemplate.TexTemplateConverter
    else:
        raise ValueError("No known converter for format '%s'." % (format))

def get_converter(format = quizgen.constants.FORMAT_JSON, **kwargs):
    converter_class = get_converter_class(format = format)
    return converter_class(**kwargs)

def convert_variant(variant, format = quizgen.constants.FORMAT_JSON,
        constructor_args = {}, converter_args = {}):
    if (not isinstance(variant, quizgen.variant.Variant)):
        raise ValueError("convert_variant() requires a quizgen.variant.Variant type, found %s." % (type(variant)))

    converter = get_converter(format = format, **constructor_args)
    return converter.convert_variant(variant, **converter_args)

def convert_question(question, format = quizgen.constants.FORMAT_JSON,
        constructor_args = {}, converter_args = {}):
    converter = get_converter(format = format, **constructor_args)
    return converter.convert_question(question, **converter_args)



"""
Convert a quiz into HTML using templates.
"""

import os

import quizgen.constants
import quizgen.converter.template

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DEFAULT_TEMPLATE_DIR = os.path.join(THIS_DIR, '..', 'data', 'templates', 'edq-html')

class HTMLTemplateConverter(quizgen.converter.template.TemplateConverter):
    def __init__(self, template_dir = DEFAULT_TEMPLATE_DIR, **kwargs):
        super().__init__(quizgen.constants.FORMAT_HTML, template_dir, **kwargs)

    def clean_solution_content(self, document):
        """
        An opportunity for children to clean the text of a solution before it is entered into a key.
        For example, tex solutions are hacky and cannot use certain functions.
        """

        return document.to_text()



import quizgen.parser
import tests.base

class TestParser(tests.base.BaseTest):
    def test_good_cases(self):
        for i in range(len(GOOD_TEST_CASES)):
            text, expected = GOOD_TEST_CASES[i]

            with self.subTest(index = i, text = text):
                document = quizgen.parser.parse_text(text)
                result = document.to_pod(include_metadata = False)

                self.assertJSONDictEqual(expected, result)

    def test_bad_cases(self):
        for i in range(len(BAD_TEST_CASES)):
            text = BAD_TEST_CASES[i]

            with self.subTest(index = i, text = text):
                try:
                    quizgen.parser.parse_text(text)
                    self.fail("Failed to raise an exception.")
                except Exception:
                    # Expected.
                    pass

# Wrap a pod parser node in a block.
def _wrap_block(nodes):
    return {
        'type': 'document',
        'nodes': [
            {
                'type': 'block',
                'nodes': nodes,
            },
        ],
    }

# Wrap text nodes in a text block.
def _wrap_text_nodes(nodes):
    return _wrap_block([{
        'type': 'text',
        'nodes': nodes,
    }])

GOOD_TEST_CASES = [
    ['Text', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Text'
        },
    ])],

    ['Foo bar', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Foo bar'
        },
    ])],

    ['\nFoo\nbar\n', _wrap_block([
        {
            'type': 'text',
            'nodes': [{
                'type': 'normal_text',
                'text': 'Foo',
            }],
        },
        {
            'type': 'text',
            'nodes': [{
                'type': 'normal_text',
                'text': 'bar',
            }],
        },
    ])],

    ['\nFoo\\nbar\n', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Foo',
        },
        {
            'type': 'linebreak',
        },
        {
            'type': 'normal_text',
            'text': 'bar'
        },
    ])],

    ['Some *italics* text.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Some '
        },
        {
            'type': 'italics_text',
            'text': 'italics'
        },
        {
            'type': 'normal_text',
            'text': ' text.'
        },
    ])],

    ['Some * spaced  italics   * text.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Some '
        },
        {
            'type': 'italics_text',
            'text': ' spaced  italics   '
        },
        {
            'type': 'normal_text',
            'text': ' text.'
        },
    ])],

    ['Some **bold** text.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Some '
        },
        {
            'type': 'bold_text',
            'text': 'bold'
        },
        {
            'type': 'normal_text',
            'text': ' text.'
        },
    ])],

    ['Some ** spaced  bold   ** text.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Some '
        },
        {
            'type': 'bold_text',
            'text': ' spaced  bold   '
        },
        {
            'type': 'normal_text',
            'text': ' text.'
        },
    ])],

    ['Escape \\\\ backslash.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape \\ backslash.'
        },
    ])],

    ['Escape \\* asterisk.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape * asterisk.'
        },
    ])],

    ['Escape \\| pipe.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape | pipe.'
        },
    ])],

    ['Escape \\` backtick.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape ` backtick.'
        },
    ])],

    ['Escape \\- dash.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape - dash.'
        },
    ])],

    ['Escape \\! bang.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape ! bang.'
        },
    ])],

    ['Escape \\[ open bracket.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape [ open bracket.'
        },
    ])],

    ['Escape \\/ slash.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Escape / slash.'
        },
    ])],

    ['`inline_code();`', _wrap_text_nodes([
        {
            'type': 'code',
            'inline': True,
            'text': 'inline_code();'
        },
    ])],

    ['Inline `code()`.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Inline '
        },
        {
            'type': 'code',
            'inline': True,
            'text': 'code()'
        },
        {
            'type': 'normal_text',
            'text': '.'
        },
    ])],

    ['`inline_code("\\`");`', _wrap_text_nodes([
        {
            'type': 'code',
            'inline': True,
            'text': 'inline_code("`");'
        },
    ])],

    ['```code_block()```', _wrap_block([
        {
            'type': 'code',
            'inline': False,
            'text': 'code_block()'
        },
    ])],

    ['``` code_block() ```', _wrap_block([
        {
            'type': 'code',
            'inline': False,
            'text': ' code_block() '
        },
    ])],

    ['```\ncode_block()\n```', _wrap_block([
        {
            'type': 'code',
            'inline': False,
            'text': 'code_block()'
        },
    ])],

    ['```foo(1, \'-2\')\nbar("|", x*);```', _wrap_block([
        {
            'type': 'code',
            'inline': False,
            'text': 'foo(1, \'-2\')\nbar("|", x*);'
        },
    ])],

    ['$ f(x) = x_i + x^2 \\alpha $', _wrap_text_nodes([
        {
            'type': 'equation',
            'inline': True,
            'text': 'f(x) = x_i + x^2 \\alpha'
        },
    ])],

    ['Inline $\\text{equation}$.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Inline '
        },
        {
            'type': 'equation',
            'inline': True,
            'text': '\\text{equation}'
        },
        {
            'type': 'normal_text',
            'text': '.'
        },
    ])],

    ['$f(\\$a)$', _wrap_text_nodes([
        {
            'type': 'equation',
            'inline': True,
            'text': 'f($a)'
        },
    ])],

    ['$$equation + block$$', _wrap_block([
        {
            'type': 'equation',
            'inline': False,
            'text': 'equation + block'
        },
    ])],

    ['$$ equation + - * / block() $$', _wrap_block([
        {
            'type': 'equation',
            'inline': False,
            'text': 'equation + - * / block()'
        },
    ])],

    ['$$\nf(x)\n$$', _wrap_block([
        {
            'type': 'equation',
            'inline': False,
            'text': 'f(x)'
        },
    ])],

    ['$$ f(x)\ng(x) $$', _wrap_block([
        {
            'type': 'equation',
            'inline': False,
            'text': 'f(x)\ng(x)'
        },
    ])],

    ['[text](url)', _wrap_text_nodes([
        {
            'type': 'link',
            'text': 'text',
            'link': 'url',
        },
    ])],

    ['![alt text](url)', _wrap_text_nodes([
        {
            'type': 'image',
            'text': 'alt text',
            'link': 'url',
        },
    ])],

    ['[text]( )', _wrap_text_nodes([
        {
            'type': 'link',
            'text': 'text',
            'link': '',
        },
    ])],

    ['[ ](url)', _wrap_text_nodes([
        {
            'type': 'link',
            'text': '',
            'link': 'url',
        },
    ])],

    ['[ some text ]( some url )', _wrap_text_nodes([
        {
            'type': 'link',
            'text': 'some text',
            'link': 'some url',
        },
    ])],

    ['[ some [\\] text ]( some (\\) url )', _wrap_text_nodes([
        {
            'type': 'link',
            'text': 'some [] text',
            'link': 'some () url',
        },
    ])],

    ['[[answerReference]]', _wrap_text_nodes([
        {
            'type': 'answer-reference',
            'text': 'answerReference',
        },
    ])],

    ['[[answer_reference]]', _wrap_text_nodes([
        {
            'type': 'answer-reference',
            'text': 'answer_reference',
        },
    ])],

    ['[[A]]', _wrap_text_nodes([
        {
            'type': 'answer-reference',
            'text': 'A',
        },
    ])],

    ['// Some comment.', _wrap_text_nodes([
        {
            'type': 'comment',
            'text': 'Some comment.'
        },
    ])],

    ['// Some // comment. \\ * | ` - ! [ / ', _wrap_text_nodes([
        {
            'type': 'comment',
            'text': 'Some // comment. \\ * | ` - ! [ /'
        },
    ])],

    ['Some // comment.', _wrap_text_nodes([
        {
            'type': 'normal_text',
            'text': 'Some '
        },
        {
            'type': 'comment',
            'text': 'comment.'
        },
    ])],

    [
        '''
| a | b | c |
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-row',
                        'head': False,
                        'cells': [
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': 'a'
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': 'b'
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': 'c'
                                    },
                                ],
                            },
                        ]
                    },
                ],
            }
        ])
    ],

    [
        '''
|---|
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-sep'
                    },
                ],
            }
        ])
    ],

    [
        '''
|---|---|
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-sep'
                    },
                ],
            }
        ])
    ],

    [
        '''
|---| |
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-sep'
                    },
                ],
            }
        ])
    ],

    [
        '''
|--- | |
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-sep'
                    },
                ],
            }
        ])
    ],

    [
        '''
|- 1  | \-2   |         3^        |
|-----|-------|-------------------|
| *a* | **b** | `c()` and $ d() $ |
        ''',
        _wrap_block([
            {
                'type': 'table',
                'rows': [
                    {
                        'type': 'table-row',
                        'head': True,
                        'cells': [
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': '1'
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': '-2',
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'normal_text',
                                        'text': '3^'
                                    },
                                ],
                            },
                        ]
                    },
                    {
                        'type': 'table-sep'
                    },
                    {
                        'type': 'table-row',
                        'head': False,
                        'cells': [
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'italics_text',
                                        'text': 'a'
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'bold_text',
                                        'text': 'b'
                                    },
                                ],
                            },
                            {
                                'type': 'text',
                                'nodes': [
                                    {
                                        'type': 'code',
                                        'inline': True,
                                        'text': 'c()'
                                    },
                                    {
                                        'type': 'normal_text',
                                        'text': ' and '
                                    },
                                    {
                                        'type': 'equation',
                                        'inline': True,
                                        'text': 'd()'
                                    },
                                ],
                            },
                        ]
                    },
                ],
            }
        ])
    ],
]

BAD_TEST_CASES = [
    '[[_]]',
    '[[1]]',
    '[[_a]]',
    '[[1a]]',
    '[[%a]]',
]



import atexit
import os
import shutil
import tempfile
import uuid

def read_if_exists(path, **kwargs):
    """
    Read the file if it exists, otherwise return None.
    """

    if (os.path.exists(path)):
        return read(path, **kwargs)

    return None

def read(path, strip = True, rstrip = True):
    with open(path, 'r') as file:
        contents = file.read()

    if (strip):
        contents = contents.strip()

    if (rstrip):
        contents = contents.rstrip()

    return contents

def write(path, contents):
    with open(path, 'w') as file:
        file.write(contents)

def get_temp_path(prefix = '', suffix = '', rm = True, mkdir = True):
    """
    Get a path to a valid temp dirent.
    mkdir will determine of this directory will exist on return.
    If rm is True, then the dirent will be attempted to be deleted on exit
    (no error will occur if the path is not there).
    """

    path = None
    while ((path is None) or os.path.exists(path)):
        path = os.path.join(tempfile.gettempdir(), prefix + str(uuid.uuid4()) + suffix)

    if (mkdir):
        os.makedirs(path, exist_ok = True)

    if (rm):
        atexit.register(remove_dirent, path)

    return path

def remove_dirent(path):
    if (not os.path.exists(path)):
        return

    if (os.path.isfile(path) or os.path.islink(path)):
        os.remove(path)
    elif (os.path.isdir(path)):
        shutil.rmtree(path)
    else:
        raise ValueError("Unknown type of dirent: '%s'." % (path))

def copy_dirent(source, dest):
    """
    Copy a file or directory into dest.
    If source is a file, then dest can be a file or dir.
    If source is a dir, then dest must be a non-existent dir.
    """

    if (os.path.isfile(source)):
        shutil.copy2(source, dest)
    else:
        shutil.copytree(source, dest)



import copy
import json
import logging
import math
import os
import random

import json5

import quizgen.common
import quizgen.constants
import quizgen.parser
import quizgen.util.file

PROMPT_FILENAME = 'prompt.md'

class Question(object):
    def __init__(self, prompt = '', question_type = '', answers = [],
            base_dir = '.',
            points = 0, base_name = '',
            shuffle_answers = True,
            custom_header = None, skip_numbering = None,
            hints = None,
            **kwargs):
        self.base_dir = base_dir

        self.prompt = prompt
        self.prompt_document = None

        self.question_type = question_type

        self.answers = answers
        self.answers_documents = None

        self.points = points
        self.base_name = base_name

        self.hints = hints

        self.shuffle_answers = shuffle_answers

        self.custom_header = custom_header
        self.skip_numbering = skip_numbering

        try:
            self.validate()
        except Exception as ex:
            raise quizgen.common.QuizValidationError(f"Error while validating question (%s)." % (self.base_dir)) from ex

    def inherit_from_group(self, group):
        """
        Inherit attributes from a group.
        """

        self.points = group.points
        self.base_name = group.name

        if (group.custom_header is not None):
            self.custom_header = group.custom_header

        if (group.skip_numbering is not None):
            self.skip_numbering = group.skip_numbering

        self.shuffle_answers = (self.shuffle_answers and group.shuffle_answers)

        for (key, value) in group.hints.items():
            # Only take hints not already set in the question.
            if (key not in self.hints):
                self.hints[key] = value

    def validate(self):
        if ((self.prompt is None) or (self.prompt == "")):
            raise quizgen.common.QuizValidationError("Prompt cannot be empty.")
        self.prompt_document = quizgen.parser.parse_text(self.prompt, base_dir = self.base_dir)

        if (self.question_type not in quizgen.constants.QUESTION_TYPES):
            raise quizgen.common.QuizValidationError(f"Unknown question type: '{self.question_type}'.")

        if (self.hints is None):
            self.hints = {}

        self._validate_answers()

    def _validate_answers(self):
        if (self.question_type in [quizgen.constants.QUESTION_TYPE_ESSAY, quizgen.constants.QUESTION_TYPE_SA]):
            self._validate_text_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_TEXT_ONLY):
            if (not isinstance(self.answers, list)):
                raise quizgen.common.QuizValidationError("Question type '%s' cannot have answers." % (self.question_type))

            if (len(self.answers) != 0):
                raise quizgen.common.QuizValidationError("Question type '%s' cannot have answers." % (self.question_type))
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_FIMB):
            self._validate_fimb_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_FITB):
            self._validate_fitb_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_MATCHING):
            self._validate_matching_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_MA):
            self._validate_answer_list()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_MCQ):
            self._validate_answer_list(min_correct = 1, max_correct = 1)
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_MDD):
            self._validate_mdd_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_NUMERICAL):
            self._validate_numerical_answers()
        elif (self.question_type == quizgen.constants.QUESTION_TYPE_TF):
            self._validate_tf_answers()
        else:
            raise quizgen.common.QuizValidationError(f"Unknown question type: '{self.question_type}'.")

    def _validate_text_answers(self):
        if (self.answers is None):
            self.answers = ['']
        elif (isinstance(self.answers, str)):
            self.answers = [self.answers]

        if (not isinstance(self.answers, list)):
            possible_answers = 'null/None, string, empty list, or list of strings'
            raise quizgen.common.QuizValidationError("Question type '%s' must an answer that is %s, found: '%s'." % (
                    self.question_type, possible_answers, str(self.answers)))

        if (len(self.answers) == 0):
            self.answers = ['']

        for i in range(len(self.answers)):
            answer = self.answers[i]
            if (not isinstance(answer, str)):
                raise quizgen.common.QuizValidationError("Question type '%s' answers must be a list of strings, element %d is '%s' (%s)." % (
                        self.question_type, i, answer, str(type(answer))))

        self.answers_documents = []
        for answer in self.answers:
            self.answers_documents.append(quizgen.parser.parse_text(answer, base_dir = self.base_dir))

    def _validate_tf_answers(self):
        if (isinstance(self.answers, bool)):
            # Change answers to look like multiple choice.
            self.answers = [
                {"correct": self.answers, "text": 'True'},
                {"correct": (not self.answers), "text": 'False'},
            ]
        elif (isinstance(self.answers, list)):
            pass
        else:
            raise quizgen.common.QuizValidationError(f"'answers' for a T/F question must be a boolean, found '{self.answers}' ({type(self.answers)}).")

        self._validate_answer_list()

        if (len(self.answers) != 2):
            raise quizgen.common.QuizValidationError("Expecting exactly two answer for T/F question, found %d." % (len(self.answers)))

        labels = list(sorted([answer['text'] for answer in self.answers]))

        expected = ['False', 'True']
        if (labels != expected):
            raise quizgen.common.QuizValidationError("T/F labels (text) not as expected. Expected: '%s', Actual: '%s'." % (expected, labels))

    def _validate_mdd_answers(self):
        if (len(self.answers) == 0):
            raise quizgen.common.QuizValidationError("No answers provided, at least one answer required.")

        self.answers_documents = {}

        for (key, answers) in self.answers.items():
            key_doc = quizgen.parser.parse_text(key, base_dir = self.base_dir)
            values_docs = _validate_answer_list(answers, self.base_dir, min_correct = 1, max_correct = 1)

            self.answers_documents[key] = {
                'key': key_doc,
                'values': values_docs,
            }

    def _validate_numerical_answers(self):
        _check_type(self.answers, list, "'answers' for a numerical question")

        for answer in self.answers:
            label = "values of 'answers' for a numerical question"
            _check_type(answer, dict, label)

            if ('type' not in answer):
                raise quizgen.common.QuizValidationError(f"Missing key ('type') for {label}.")

            if (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_EXACT):
                required_keys = ['value']

                if ('margin' not in answer):
                    answer['margin'] = 0
            elif (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_RANGE):
                required_keys = ['min', 'max']
            elif (answer['type'] == quizgen.constants.NUMERICAL_ANSWER_TYPE_PRECISION):
                required_keys = ['value', 'precision']
            else:
                raise quizgen.common.QuizValidationError(f"Unknown numerical answer type: '{answer['type']}'.")

            for key in required_keys:
                if (key not in answer):
                    raise quizgen.common.QuizValidationError(f"Missing required key '{key}' for numerical answer type '{answer['type']}'.")

    def _validate_fitb_answers(self):
        """
        Just set up the answers to look like fill in multiple blanks with an empty key.
        """

        label = "fill in the blank"

        # Standardize an alternate format.
        # Note that this is the format we will eventually output,
        # but we want to validate the user-facing format so errors are more clear.
        if (isinstance(self.answers, dict)):
            if (len(self.answers) != 1):
                raise quizgen.common.QuizValidationError("Dict format should have exactly one entry (''), found %d entries." % (len(self.answers)))

            if ('' not in self.answers):
                raise quizgen.common.QuizValidationError("Dict format does not have required key '' (empty string).")

            self.answers = self.answers['']

        _check_type(self.answers, list, f"{label} answers")

        if (len(self.answers) == 0):
            raise quizgen.common.QuizValidationError(f"Expected {label} answers to be non-empty.")

        self.answers = {"": self.answers}

        self._validate_fimb_answers(label = label)

    def _validate_fimb_answers(self, label = 'fill in multiple blanks'):
        if (not isinstance(self.answers, dict)):
            raise quizgen.common.QuizValidationError(f"Expected dict for {label} answers, found {type(self.answers)}.")

        if (len(self.answers) == 0):
            raise quizgen.common.QuizValidationError(f"Expected {label} answers to be non-empty.")

        for (key, values) in self.answers.items():
            if (not isinstance(values, list)):
                self.answers[key] = [values]

        for (key, values) in self.answers.items():
            _check_type(key, str, f"key for {label} answers")

            if (len(values) == 0):
                raise quizgen.common.QuizValidationError(f"Expected {label} possible values to be non-empty.")

            for value in values:
                _check_type(value, str, f"value for {label} answers")

        self.answers_documents = {}
        for (key, values) in self.answers.items():
            key_doc = quizgen.parser.parse_text(key, base_dir = self.base_dir)

            values_docs = []
            for value in values:
                values_docs.append(quizgen.parser.parse_text(value, base_dir = self.base_dir))

            self.answers_documents[key] = {
                'key': key_doc,
                'values': values_docs,
            }

    def _validate_matching_answers(self):
        if (not isinstance(self.answers, dict)):
            raise quizgen.common.QuizValidationError(f"Expected dict for matching answers, found {type(self.answers)}.")

        if ('matches' not in self.answers):
            raise quizgen.common.QuizValidationError("Matching answer type is missing the 'matches' field.")

        for match in self.answers['matches']:
            if (len(match) != 2):
                raise quizgen.common.QuizValidationError(f"Expected exactly two items for a match list, found {len(match)}.")

        if ('distractors' not in self.answers):
            self.answers['distractors'] = []

        for i in range(len(self.answers['distractors'])):
            distractor = self.answers['distractors'][i]
            if (not isinstance(distractor, str)):
                raise quizgen.common.QuizValidationError(f"Distractors must be strings, found {type(distractor)}.")

            distractor = distractor.strip()

            if ("\n" in distractor):
                raise quizgen.common.QuizValidationError(f"Distractors cannot have newlines, found {type(distractor)}.")

            self.answers['distractors'][i] = distractor

        self.answers_documents = {
            'matches': [],
            'distractors': [],
        }

        for (left, right) in self.answers['matches']:
            left_doc = quizgen.parser.parse_text(left, base_dir = self.base_dir)
            right_doc = quizgen.parser.parse_text(right, base_dir = self.base_dir)

            self.answers_documents['matches'].append([left_doc, right_doc])

        for distractor in self.answers['distractors']:
            doc = quizgen.parser.parse_text(distractor, base_dir = self.base_dir)
            self.answers_documents['distractors'].append(doc)

    def _validate_answer_list(self, min_correct = 0, max_correct = math.inf):
        self.answers_documents = _validate_answer_list(self.answers, self.base_dir,
                min_correct = min_correct, max_correct = max_correct)

    def to_json(self, indent = 4, include_docs = True):
        return json.dumps(self.to_dict(include_docs = include_docs), indent = indent)

    def to_dict(self, include_docs = True):
        value = self.__dict__.copy()

        value['answers'] = self._answers_to_dict(self.answers)

        if (include_docs):
            value['prompt_document'] = self.prompt_document.to_pod()
            value['answers_documents'] = self._answers_to_dict(self.answers_documents)
        else:
            value.pop('prompt_document', None)
            value.pop('answers_documents', None)

        return value

    @staticmethod
    def from_dict(data, base_dir = None):
        data = data.copy()

        if (base_dir is not None):
            data['base_dir'] = base_dir
        elif ('base_dir' not in data):
            data['base_dir'] = '.'

        data.pop('prompt_document', None)
        data.pop('answers_documents', None)

        return Question(**data)

    def _answers_to_dict(self, target):
        if (isinstance(target, dict)):
            return {key: self._answers_to_dict(value) for (key, value) in target.items()}
        elif (isinstance(target, list)):
            return [self._answers_to_dict(answer) for answer in target]
        elif (isinstance(target, quizgen.parser.ParseNode)):
            return target.to_pod()
        else:
            return target

    def collect_file_paths(self):
        paths = []

        paths += self.prompt_document.collect_file_paths(self.base_dir)

        for document in self._collect_documents(self.answers):
            paths += document.collect_file_paths(self.base_dir)

        return paths

    def _collect_documents(self, target):
        if (isinstance(target, dict)):
            return self._collect_documents(list(target.values()))
        elif (isinstance(target, list)):
            documents = []
            for value in target:
                documents += self._collect_documents(value)
            return documents
        elif (isinstance(target, quizgen.parser.ParseNode)):
            return [target]
        else:
            return []

    def should_skip_numbering(self):
        return ((self.skip_numbering is not None) and (self.skip_numbering))

    @staticmethod
    def from_path(path):
        with open(path, 'r') as file:
            question_info = json5.load(file)

        # Check for a prompt file.
        prompt_path = os.path.join(os.path.dirname(path), PROMPT_FILENAME)
        if (os.path.exists(prompt_path)):
            question_info['prompt'] = quizgen.util.file.read(prompt_path)
            logging.debug("Loading question prompt from '%s'.", prompt_path)

        base_dir = os.path.dirname(path)

        return Question(base_dir = base_dir, **question_info)

    def copy(self):
        return copy.deepcopy(self)

    def shuffle(self, rng = None):
        if (not self.shuffle_answers):
            return

        if (rng is None):
            rng = random.Random()

        # Nothing to be done for these types.
        skip_types = [
            quizgen.constants.QUESTION_TYPE_ESSAY,
            quizgen.constants.QUESTION_TYPE_FIMB,
            quizgen.constants.QUESTION_TYPE_FITB,
            quizgen.constants.QUESTION_TYPE_NUMERICAL,
            quizgen.constants.QUESTION_TYPE_SA,
            quizgen.constants.QUESTION_TYPE_TEXT_ONLY,
        ]

        if (self.question_type in skip_types):
            return

        # List types that just need a simple shuffle.
        list_types = [
            quizgen.constants.QUESTION_TYPE_MA,
            quizgen.constants.QUESTION_TYPE_MCQ,
            quizgen.constants.QUESTION_TYPE_TF,
        ]

        if (self.question_type in list_types):
            collection = list(zip(self.answers, self.answers_documents))
            rng.shuffle(collection)
            self.answers, self.answers_documents = map(list, zip(*collection))
            return

        if (self.question_type == quizgen.constants.QUESTION_TYPE_MATCHING):
            # Matching is special because it requires additional shuffling support at the converter level.
            self.answers['shuffle'] = True
            self.answers['shuffle_seed'] = rng.randint(0, 2 ** 64)
            return

        if (self.question_type == quizgen.constants.QUESTION_TYPE_MDD):
            for key in self.answers:
                collection = list(zip(self.answers[key], self.answers_documents[key]['values']))
                rng.shuffle(collection)
                self.answers[key], self.answers_documents[key]['values'] = map(list, zip(*collection))

            return

        raise ValueError("Unknown question type: '%s'." % (self.question_type))

def _check_type(value, expected_type, label):
    if (not isinstance(value, expected_type)):
        raise quizgen.common.QuizValidationError(f"{label} must be a {expected_type}, found '{value}' ({type(value)}).")

def _validate_answer_list(answers, base_dir, min_correct = 0, max_correct = math.inf):
    _check_type(answers, list, "'answers'")

    if (len(answers) == 0):
        raise quizgen.common.QuizValidationError(f"No answers provided, at least one answer required.")

    num_correct = 0
    for answer in answers:
        if ('correct' not in answer):
            raise quizgen.common.QuizValidationError(f"Answer has no 'correct' field (base_dir: '{base_dir}'.")

        if ('text' not in answer):
            raise quizgen.common.QuizValidationError(f"Answer has no 'text' field (base_dir: '{base_dir}'.")

        if (answer['correct']):
            num_correct += 1

    if (num_correct < min_correct):
        raise quizgen.common.QuizValidationError(f"Did not find enough correct answers. Expected at least {min_correct}, found {num_correct} (base_dir: '{base_dir}'.")

    if (num_correct > max_correct):
        raise quizgen.common.QuizValidationError(f"Found too many correct answers. Expected at most {max_correct}, found {num_correct} (base_dir: '{base_dir}'.")

    answers_documents = []
    for answer in answers:
        doc = quizgen.parser.parse_text(answer['text'], base_dir = base_dir)
        answers_documents.append(doc)

    return answers_documents



"""
Convert a quiz to JSON.
"""

import quizgen.converter.converter

class JSONConverter(quizgen.converter.converter.Converter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def convert_variant(self, variant, **kwargs):
        return variant.to_json()



import glob
import os

import quizgen.common
import quizgen.question

QUESTION_FILENAME = 'question.json'

class Group(object):
    def __init__(self, name = '',
            pick_count = 1, points = 10,
            shuffle_answers = True,
            custom_header = None, skip_numbering = None,
            hints = None,
            questions = [], **kwargs):
        self.name = name
        self.pick_count = pick_count
        self.points = points

        self.shuffle_answers = shuffle_answers

        self.custom_header = custom_header
        self.skip_numbering = skip_numbering

        self.hints = hints

        self.questions = questions

        try:
            self.validate()
        except Exception as ex:
            raise quizgen.common.QuizValidationError(f"Error while validating group (%s)." % (self.name)) from ex

    def validate(self):
        if ((self.name is None) or (self.name == "")):
            raise quizgen.common.QuizValidationError("Name cannot be empty.")

        if (self.pick_count < 0):
            raise quizgen.common.QuizValidationError("Pick count cannot be negative.")

        if (self.hints is None):
            self.hints = {}

        if (not isinstance(self.questions, list)):
            raise quizgen.common.QuizValidationError("Questions must be a non-empty list, found: '%s'." % (str(self.questions)))

        if (len(self.questions) == 0):
            raise quizgen.common.QuizValidationError("Questions must be non-empty.")

        for question in self.questions:
            question.inherit_from_group(self)

    def to_dict(self, include_docs = True):
        value = self.__dict__.copy()
        value['questions'] = [question.to_dict(include_docs = include_docs) for question in self.questions]
        return value

    def collect_file_paths(self):
        paths = []

        for question in self.questions:
            paths += question.collect_file_paths()

        return paths

    @staticmethod
    def from_dict(group_info, base_dir):
        group_info = group_info.copy()

        paths = []
        for path in group_info.get('questions', []):
            if (not os.path.isabs(path)):
                path = os.path.join(base_dir, path)
            paths.append(os.path.abspath(path))

        paths = list(sorted(set(paths)))

        questions = []
        for path in paths:
            questions += _parse_questions(path)

        group_info['questions'] = questions

        return Group(**group_info)

    def copy_questions(self):
        return [question.copy() for question in self.questions]

    def choose_questions(self, rng):
        return [question.copy() for question in rng.sample(self.questions, k = self.pick_count)]

def _parse_questions(path):
    if (not os.path.exists(path)):
        raise quizgen.common.QuizValidationError(f"Question path does not exist: '{path}'.")

    if (os.path.isfile(path)):
        return [quizgen.question.Question.from_path(path)]

    questions = []
    for subpath in sorted(glob.glob(os.path.join(path, '**', QUESTION_FILENAME), recursive = True)):
        questions.append(quizgen.question.Question.from_path(subpath))

    return questions



import logging
import shutil
import subprocess

ENCODING = 'utf-8'

def _has_command(command, cwd = '.'):
    result = subprocess.run(["which", command], cwd = cwd, capture_output = True)
    return (result.returncode == 0)

def _has_package(package, cwd = '.'):
    result = subprocess.run(["npm", "list", package], cwd = cwd, capture_output = True)
    return (result.returncode == 0)

def is_available(cwd = '.'):
    if (shutil.which('npx') is None):
        logging.warning("Could not find `npx` (usually installed with `npm`), cannot use katex equations.")
        return False

    if (not _has_package('katex', cwd = cwd)):
        logging.warning("Could not find the `katex` package, cannot use katex equations.")
        return False

    return True

def to_html(text, cwd = '.'):
    result = subprocess.run(["npx", "katex", "--format", "mathml"], cwd = cwd,
        input = text, encoding = ENCODING,
        capture_output = True)

    if (result.returncode != 0):
        raise ValueError("KaTeX did not exit cleanly. Stdout: '%s', Stderr: '%s'" % (result.stdout, result.stderr))

    return result.stdout



import os

import quizgen.util.file
import quizgen.util.git
import tests.base

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

class TestGit(tests.base.BaseTest):
    def test_in_repo(self):
        # This test should live inside a repo.
        version = quizgen.util.git.get_version(THIS_DIR)
        self.assertNotEqual(quizgen.util.git.UNKNOWN_VERSION, version, 'Got an unknown version. (This can also happen if this instance of the quizgen project is not in a git repo.')

    def test_cwd(self):
        # The tests should be run in a repo (this project repo).
        version = quizgen.util.git.get_version()
        self.assertNotEqual(quizgen.util.git.UNKNOWN_VERSION, version, 'Got an unknown version. (This can also happen if the tests are not being run inside of a git repo.')

    def test_not_in_repo(self):
        # A new temp dir should not be in a git repo.
        path = quizgen.util.file.get_temp_path(prefix = 'quizgen-test-git-')
        version = quizgen.util.git.get_version(path)
        self.assertEqual(quizgen.util.git.UNKNOWN_VERSION, version)
